import os
import time
import torch
import wandb

import numpy as np
import pandas as pd
import torch.nn.functional as F

from torch.utils.data import DataLoader
from tqdm import tqdm


import sys
sys.path.append('../')
import dinomn.mnds as mnds
import dinomn.detection as detection

        
class DiceLoss(torch.nn.Module):
    def __init__(self, alpha=0.8, beta=0.2, smoothing=1e-5, reduction='mean'):
        """_summary_

        Args:
            alpha (_type_): weight for micronuclei class, default=0.8
            beta (_type_): weight for nuclei class, default=0.2
            smoothing (_type_, optional): smoothing parameter for numerical stability. Defaults to 1e-5.
            reduction (str, optional): Reduction method. Defaults to 'mean'.
        """
        super(DiceLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.smoothing = smoothing
        self.reduction = reduction

    def forward(self, prediction, ground_truth):
        # Ground truth expected to be binaries, 0s and 1s
        # Conclusion, do not use one-hot encoding
        
        assert prediction.shape == ground_truth.shape, f'Predictions shape does not match the ground truth!'
        
        probs = torch.sigmoid(prediction)
        ground_truth = ground_truth.long() # convert torch.tensor 3.14 to 3, etc.
        
        num = probs * ground_truth # numerator
        num = torch.sum(num, dim=(2,3))  # Sum over all pixels NxCxHxW --> NxC
        
        den1 = probs * probs # 1st denominator
        den1 = torch.sum(den1, dim=(2,3))
        
        den2 = ground_truth * ground_truth # 2nd denominator
        den2 = torch.sum(den2, dim=(2,3))
        
        dice_loss_mn = 2. * (num[:,0]+ self.smoothing) / (den1[:,0] + den2[:,0] + self.smoothing)
        dice_loss_n = 2. * (num[:,1]+ self.smoothing) / (den1[:,1] + den2[:,1] + self.smoothing)
        
        if self.reduction == 'mean':
            dice_loss = 1 - (self.alpha * torch.mean(dice_loss_mn) + self.beta * torch.mean(dice_loss_n))
            # dice_loss = 1 - torch.mean(dice_loss)
        elif self.reduction == 'sum':
            dice_loss = 1 - (self.alpha * torch.sum(dice_loss_mn) + self.beta * torch.sum(dice_loss_n))
            # dice_loss = 1 - torch.sum(dice_loss)
        else:
            raise ValueError("'Reduction method must be either 'mean' or 'sum'")
        
        return dice_loss
    
class FocalLoss(torch.nn.Module):
    """_summary_
    Code are copied from torchvision.ops.sigmoid_focal_loss function
    """
    def __init__(self, alpha=0.25, gamma=2.0, reduction: str="mean"):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction

    def forward(self, inputs, targets):
        ce_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")
        p = torch.sigmoid(inputs)
        p_t = p * targets + (1 - p) * (1 - targets)
        loss = ce_loss * ((1 - p_t) ** self.gamma)

        if self.alpha >= 0:
            alpha_t = self.alpha * targets + (1 - self.alpha) * (1 - targets)
            loss = alpha_t * loss
            
        if self.reduction == "mean":
            loss = loss.mean()
        elif self.reduction == "sum":
            loss = loss.sum()
        else:
            raise ValueError(
                f"Invalid Value for arg 'reduction': '{self.reduction} \n Supported reduction modes: 'mean', 'sum'"
            )
        return loss
    
class CombinedFocalDiceLoss(torch.nn.Module):
    def __init__(self, focal_weight=0.95, dice_weight=0.05, alpha=0.25, gamma=2.0, reduction='mean', dice_alpha=0.8, dice_beta=0.2, smoothing=1e-5):
        super().__init__() # call initialization method from torch.nn.Module
        self.focal_loss = FocalLoss(alpha=alpha, gamma=gamma, reduction=reduction)
        self.dice_loss = DiceLoss(alpha=dice_alpha, beta=dice_beta, smoothing=smoothing, reduction=reduction)
        self.focal_weight = focal_weight
        self.dice_weight = dice_weight

    def forward(self, inputs, targets):
        loss_focal = self.focal_loss(inputs, targets)
        loss_dice = self.dice_loss(inputs, targets)
        combined_loss = self.focal_weight * loss_focal + self.dice_weight * loss_dice
        return combined_loss
        
        

class MicronucleiModel(torch.nn.Module):
    # repo_url = "yifanren/DinoMN"
    # pipeline_tag = "DinoMN-Model"
    # license = "mit"
    
    def __init__(self, device, data_dir='', edges=False, patch_size=256, scale_factor=1.0, gaussian=True):
        super().__init__()
        
        self.data_dir = data_dir
        self.device = device
        self.edges = edges
        self.patch_size = patch_size
        self.scale_factor = scale_factor
        self.gaussian = gaussian
        
        self.train_dir = os.path.join(data_dir, 'train')
        self.val_dir = os.path.join(data_dir, 'validation')
        

        
    def start_model(self, batch_size, learning_rate, loss_fn, finetune=False, weight_decay=1e-6):
        if len(os.listdir(self.train_dir)) > 0:
            self.training_set = mnds.MicronucleiDataset(
                directory=self.train_dir, 
                mode="random",
                edges=self.edges,
                transform=mnds.detection_transforms,
                scale_factor=self.scale_factor,
                patch_size=self.patch_size,
                gaussian=self.gaussian
            )
        else:
            raise OSError(f"Training directory is empty: '{self.train_dir}'")
        
        if len(os.listdir(self.val_dir)) > 0:
            self.validation_set = mnds.MicronucleiDataset(
                directory=self.val_dir,
                mode="fixed",
                edges=self.edges,
                scale_factor=self.scale_factor,
                patch_size=self.patch_size,
                gaussian=self.gaussian
            )
        else:
            raise OSError(f"Validation directory is empty: '{self.val_dir}'")
       
        self.train_dataloader = DataLoader(self.training_set, batch_size=batch_size, shuffle=True)
        self.val_dataloader = DataLoader(self.validation_set, batch_size=4, shuffle=False)
        
        self.model = detection.DetectionModel(device=self.device, finetune=finetune)
    
        # self.loss_fn = torch.nn.BCEWithLogitsLoss()
        if loss_fn == 'dice':
            self.loss_fn = DiceLoss(alpha=0.8, beta=0.2, smoothing=1e-5, reduction='mean')
        elif loss_fn == 'focal':
            self.loss_fn = FocalLoss(alpha=0.25, gamma=1, reduction='mean')
        elif loss_fn == 'combined':
            # Use all default parameters, gamma = 2 so far is good
            self.loss_fn = CombinedFocalDiceLoss(focal_weight=0.95, dice_weight=0.05, alpha=0.25, gamma=2, reduction='mean', dice_alpha=0.8, dice_beta=0.2, smoothing=1e-5)
            
        self.optimizer = torch.optim.AdamW(self.model.parameters(), lr=learning_rate, weight_decay=weight_decay) #, momentum=0.9)
        
        self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer=self.optimizer,
            T_max=20, # max number of training epochs,
            eta_min=learning_rate * 0.1
        )
        
        
    def train_one_epoch(self):
        running_loss = 0.
        last_loss = 0.

        # For each training epoch, randomize the coordinates to reduce overfitting
        # Avoid that model sees the same regions in each training epoch
        self.train_dataloader.dataset.randomize_patch_index()
        for i, data in enumerate(self.train_dataloader):
            x, y = data
            
            self.optimizer.zero_grad()
            p = self.model(x.to(self.device))

            # output resolution: 128, interpolate to 256
            p = torch.nn.functional.interpolate(p, (self.patch_size, self.patch_size))
            
            # Loss function   
            Y = (y.to(self.device) > 0).float() # convert to binary (0 & 1)
            loss = self.loss_fn(p, Y)
            
            # Training instructions
            loss.backward()
            
            # Update weights
            self.optimizer.step()
            
            # Report results
            running_loss += loss.item()
        return running_loss / (i+1)
    
    
    def train(self, epochs, batch_size, learning_rate, loss_fn, finetune=False, weight_decay=1e-6, wandb_mode=False):
        self.start_model(batch_size=batch_size, learning_rate=learning_rate, loss_fn=loss_fn, finetune=finetune, weight_decay=weight_decay)
        
        best_vloss = 1_000_000.

        start = time.time()
        for epoch in range(epochs):
            # Training
            # print(f'EPOCH {epoch} - ', end='') # comment only for grid search purpose
            T = time.time()
            self.model.train(True)
            avg_loss = self.train_one_epoch()
            
            # Update learning rate per epoch
            if wandb_mode:
                wandb.log({'current_lr':self.optimizer.param_groups[0]['lr']})
            self.scheduler.step()
            

            # Validation
            running_vloss = 0.0
            self.model.eval()
            with torch.no_grad():
                for i, vdata in enumerate(self.val_dataloader):
                    vin, vls = vdata
                    vout = self.model(vin.to(self.device))
                    # output resolution: 128
                    vout = torch.nn.functional.interpolate(vout, (self.patch_size, self.patch_size))
                    Y = (vls.to(self.device) > 0).float() # convert to binary (0 & 1)
                    
                    vloss = self.loss_fn(vout, Y)
                    running_vloss += vloss
            avg_vloss = running_vloss / (i+1)
            C = time.time() - T
            print(f'LOSS: Training: {avg_loss} - Validation: {avg_vloss} - Time: {C:.2f} secs') # comment only for grid search purpose

            # log metrics to wandb
            if wandb_mode:
                wandb.log({"Train_loss":avg_loss, "Validation_loss":avg_vloss}, commit=False) # avoid logging more steps than num of epochs

        C = time.time() - start
        print(f"\nTrainined finished in {C:.2f} seconds") # comment out for grid search
        if wandb_mode:
            wandb.log({"Train time":C})

        
    def save(self, outdir="models/", model_name='model'):
        output_file = f'{self.data_dir}{outdir}{model_name}.pth'
        torch.save(self.model.state_dict(), output_file)

        
    def load(self, model_path):
        self.model = detection.DetectionModel(device=self.device)
        self.model.load_state_dict(torch.load(model_path))
        self.model.to(self.device)
        
    
    def _generate_fixed_coord(self, im, step, patch_size):
        C,H,W = im.shape
        X = list(range(0,W-patch_size+1, step))
        Y = list(range(0,H-patch_size+1, step))
        patches_per_image = len(X) * len(Y)
        X,Y = np.meshgrid(X,Y, indexing='ij')
        X = X.reshape((patches_per_image,))
        Y = Y.reshape((patches_per_image,))
        coord = np.stack((Y,X)).T
        
        return coord
        
        
    def predict(self, image, stride=1, step=16, batch_size=512):
        classes = self.model.classifier.out_channels
        probabilities = np.zeros((classes, image.shape[0]//stride, image.shape[1]//stride), dtype=np.float32)
        counts = np.zeros((image.shape[0]//stride, image.shape[1]//stride), dtype=np.float32)
        TOKENS_PER_PATCH = self.patch_size // stride # 256, same as self.patch_size
        ones = np.ones((TOKENS_PER_PATCH, TOKENS_PER_PATCH))
        batch, coords = [], []

        self.model.eval() # finetune not allowed, use pre-trained weights!

        def batch_predict(batch, coords):
            B = torch.cat(batch, axis=0)
            # pred0 = F.softmax(self.model(B.to(self.device))) need to be changed
            output = self.model(B.to(self.device))
            
            output = torch.nn.functional.interpolate(output, (self.patch_size,self.patch_size))
            
            # the output is not probability here (weights instead), last layer of detection model is binary classification, so we transform output to binary values.
            # classify pixel-wisely to 0 or 1, and scan all over the input image, and then average them, it is probability at the end.
            output = output > 0.0
            
            pred0 = output.float()
            P = torch.reshape(pred0, (-1, classes, TOKENS_PER_PATCH, TOKENS_PER_PATCH))
            # P.detach().cpu().numpy() might be better
            P = P.cpu().numpy()
            # print(f'P shape in batch_predict(): {P.shape}')
            
            for c in range(len(coords)):
                y = coords[c]["a"]
                x = coords[c]["b"]
                probabilities[:,y:y+TOKENS_PER_PATCH,x:x+TOKENS_PER_PATCH] += P[c]
                counts[y:y+TOKENS_PER_PATCH,x:x+TOKENS_PER_PATCH] += ones
            coords = []
        
        with torch.no_grad(): # turn off gradients for inference
            image = mnds.patch_to_rgb(image)
            C = self._generate_fixed_coord(im=image, step=step, patch_size=self.patch_size)
            
            for i in range(len(C)): # loop through coordinates
                y,x = C[i,:]
                coords.append({'a':y, 'b':x})
                
                crop = image[:,y:y+self.patch_size,x:x+self.patch_size]
                batch.append(crop[None,:,:,:])
                
                if len(batch) == batch_size:
                        # Get predictions
                        batch_predict(batch, coords)
                        batch, coords = [], []
            
            # remaining batches that is lower than specified batch size
            if len(batch) > 0:
                batch_predict(batch, coords)
                batch, coords = [], []

        probabilities = probabilities/counts
        return probabilities
    


import torch
import torch.nn.functional as F
# import vision_transformer as vit


# class TruncViT(vit.VisionTransformer):
    
#     def __init__(self, img_size=[224], patch_size=16, in_chans=3, num_classes=0, embed_dim=768, depth=12,
#                  num_heads=12, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop_rate=0., attn_drop_rate=0.,
#                  drop_path_rate=0., norm_layer=torch.nn.LayerNorm, **kwargs):
#         super().__init__(img_size, patch_size, in_chans, num_classes, embed_dim, depth,
#                  num_heads, mlp_ratio, qkv_bias, qk_scale, drop_rate, attn_drop_rate,
#                  drop_path_rate, norm_layer)
#         del(self.head)

#     def forward(self, x):
#         x = self.prepare_tokens(x)
#         for blk in self.blocks:
#             x = blk(x)
#         x = self.norm(x)
#         return x#[:, 0]
    
# def trunc_vit_tiny(patch_size=16, **kwargs):
#     model = TruncViT(
#         patch_size=patch_size, embed_dim=192, depth=12, num_heads=3, mlp_ratio=4,
#         qkv_bias=True, norm_layer=vit.partial(torch.nn.LayerNorm, eps=1e-6), **kwargs)
#     return model

class DetectionModel(torch.nn.Module):
    
    def __init__(self, device, stride=8, finetune=False):
        super().__init__()
        
        self.finetune = finetune
        
        # pretrained backbone has patch size 14 x 14, split into 14 row and columns
        self.feature_extractor = torch.hub.load('facebookresearch/dinov2', 'dinov2_vits14_reg').to(device) # dinov2 vit small model

        def conv_block(in_channels, out_channels, kernel_size=(3,3), padding=(1,1), norm_shape=[96, 128, 128]):
            return torch.nn.Sequential(
                torch.nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, padding=padding),
                torch.nn.LayerNorm(norm_shape),
                torch.nn.ReLU()
            )
            
        self.upscale1 = torch.nn.ConvTranspose2d(in_channels=384, out_channels=192, kernel_size=(2,2), stride=2)
        self.upscale1.to(device)
        self.block1 = conv_block(in_channels=192, out_channels=192, kernel_size=(3,3), padding=(1,1), norm_shape=[192, 64, 64])
        self.block2 = conv_block(in_channels=192, out_channels=192, kernel_size=(3,3), padding=(1,1), norm_shape=[192, 64, 64])
        self.block3 = conv_block(in_channels=192, out_channels=192, kernel_size=(3,3), padding=(1,1), norm_shape=[192, 64, 64])
        
        self.block1.to(device)
        self.block2.to(device)
        self.block3.to(device)
        
        # here number of features (d_model) represents channels
        self.decoder_layer1 = torch.nn.TransformerDecoderLayer(d_model=192, nhead=8, activation='relu', batch_first=True)  # Assuming d_model is 192
        self.transformer_decoder1 = torch.nn.TransformerDecoder(self.decoder_layer1, num_layers=4)
        self.transformer_decoder1.to(device)
        
        self.projection1 = torch.nn.Linear(384, 192)
        self.projection1.to(device)
        
        self.upscale2 = torch.nn.ConvTranspose2d(in_channels=192, out_channels=96, kernel_size=(2,2), stride=2)
        self.upscale2.to(device)
        self.block4 = conv_block(in_channels=96, out_channels=96, kernel_size=(3,3), padding=(1,1), norm_shape=[96, 128, 128])
        self.block5 = conv_block(in_channels=96, out_channels=96, kernel_size=(3,3), padding=(1,1), norm_shape=[96, 128, 128])
        self.block6 = conv_block(in_channels=96, out_channels=96, kernel_size=(3,3), padding=(1,1), norm_shape=[96, 128, 128])
        
        self.block4.to(device)
        self.block5.to(device)
        self.block6.to(device)
        
        self.projection2 = torch.nn.Linear(192, 96)
        self.projection2.to(device)
        
        # classification layer
        self.classifier = torch.nn.Conv2d(in_channels=96, out_channels=2, kernel_size=(1,1))
        self.classifier.to(device)
                
    def forward(self, x):
        # Skip Connections
        if self.finetune:
            x = torch.nn.functional.interpolate(x, (448,448))
            x = self.feature_extractor.forward_features(x)['x_norm_patchtokens']
            B,T,C = x.shape # Batch, Token size * Toekn size, Channel
            H,W = 32,32
            
            memory = x # original image features 1,1024,384
            x = x.reshape(B,H,W,C).permute(0,3,1,2) # 1,384,32,32
            
            # Pixel Decoder Block 1
            x = self.upscale1(x)
            residual1 = x
            x = self.block1(x)
            residual2 = x
            x = self.block2(x) + residual1
            x = self.block3(x) + residual2 # x: 1,192,64,64
            
            # Transformer Decoder Block 1
            B,C,H,W = x.shape
            target = x.reshape(B,H*W,C)
            memory = self.projection1(memory)
            transformer_x = self.transformer_decoder1(tgt=target, memory=memory)
            # 1,4096,192
            
            # Pixel Decoder Block 2
            x = self.upscale2(x)
            residual3 = x
            x = self.block4(x)
            residual4 = x
            x = self.block5(x) + residual3
            x = self.block6(x) + residual4 # x: 1,96,128,128
            
            
            transformer_x = self.projection2(transformer_x)
            transformer_x = transformer_x.reshape(B,96,H,W)
            transformer_x = F.interpolate(x, (128,128))
            
            # Sum up information
            x = x + transformer_x
            
            x = self.classifier(x)
        else:
            with torch.no_grad():
                x = torch.nn.functional.interpolate(x, (448,448))
                x = self.feature_extractor.forward_features(x)['x_norm_patchtokens']
            B,T,C = x.shape # Batch, Token size * Toekn size, Channel
            H,W = 32,32
            memory = x # original image features
            x = x.reshape(B,H,W,C).permute(0,3,1,2) # 1,384,32,32
  
            # Pixel Decoder Block 1
            x = self.upscale1(x)
            residual1 = x
            x = self.block1(x)
            residual2 = x
            x = self.block2(x) + residual1
            x = self.block3(x) + residual2 # x: 1,192,64,64
            
            # Transformer Decoder Block 1
            B,C,H,W = x.shape
            target = x.reshape(B,H*W,C)
            memory = self.projection1(memory)
            transformer_x = self.transformer_decoder1(tgt=target, memory=memory)
            
            # Pixel Decoder Block 2
            x = self.upscale2(x)
            residual3 = x
            x = self.block4(x)
            residual4 = x
            x = self.block5(x) + residual3
            x = self.block6(x) + residual4 # x: 1,96,128,128
            
            transformer_x = self.projection2(transformer_x)
            transformer_x = transformer_x.reshape(B,96,H,W)
            transformer_x = F.interpolate(x, (128,128))
            
            # Sum up information
            x = x + transformer_x
            
            x = self.classifier(x)
        
        return x



import os
import skimage
import random
import torch
import numpy as np
import pandas as pd

from tqdm import tqdm
from torch.utils.data import Dataset
import torchvision.transforms.functional as TF

# GRAY SCALE PATCH TO RGB IMAGE
def patch_to_rgb(patch, edges=False):
    if edges:
        sobel = skimage.filters.sobel(patch)
        sobel = 2*skimage.exposure.rescale_intensity(sobel, out_range=np.float32)
        sobel[sobel > 1.] = 1.
    else:
        sobel = patch
        
    px = np.concatenate(
        (sobel[np.newaxis,:,:], patch[np.newaxis,:,:], patch[np.newaxis,:,:]), 
        axis=0)
    return torch.Tensor(px)

# OPEN AN IMAGE
def read_image(image_path, scale=1.0):
    im = skimage.io.imread(image_path)
    if scale != 1.:
        im = skimage.transform.rescale(im, scale)
    return im


# READ MICRONUCLEI ANNOTATIONS
def read_micronuclei_annotations(directory, filename, size_filter=1e9, scale_factor=1.0):
    mask = read_image(os.path.join(directory, 'nuclei_masks', filename), scale=scale_factor)
    img = read_image(os.path.join(directory, 'images', filename), scale=scale_factor)
    
    imid = filename.split('.')[0]
    labels = skimage.measure.label(mask)
    data = []
    for i in range(1,len(np.unique(labels))):
        ys,xs = np.where(labels == i)
        intensity = np.mean(img[ys,xs])
        a,b = int(np.mean(ys)), int(np.mean(xs))
        area = np.sum(labels == i)
        if area <= size_filter:
            a = int(scale_factor * a)
            b = int(scale_factor * b)
            data.append({"Image":f'{directory}/{imid}.phenotype.tif', "x":b, "y":a, "area":area, "intensity": intensity})

    mni = pd.DataFrame(data=data, columns=["Image","x","y","area","intensity"])
    return mni #, labels


def read_nuclei_masks(directory, filename, scale_factor=1.0):
    # otl = read_image(directory, imid, 'nuclei.tif', scale=scale_factor)
    otl = read_image(os.path.join(directory, 'nuclei_masks', filename), scale=scale_factor)
    otl = otl > 0 # It's a labeled matrix, so make it binary
    return otl

# PATCH AUGMENTATIONS
def detection_transforms(patch, target):
    # Rotations
    if random.random() > 0.25:
        angle = random.choice([90, 180, 270])
        patch = TF.rotate(patch, angle)
        target = TF.rotate(target, angle)
    
    # Horizontal flips
    if random.random() > 0.5:
        patch = TF.hflip(patch)
        target = TF.hflip(target)
    
    # Brightness adjustments
    if random.random() > 0.5:
        brightness = min(2., max(0.5, np.random.normal(1, 0.5)))
        patch = TF.adjust_brightness(patch, brightness)
       
    # Contrast adjustments
    if random.random() > 0.5:
        contrast = min(2., max(0.5, np.random.normal(1, 0.5)))
        patch = TF.adjust_contrast(patch, contrast)
    
    return patch, target

# DATASET CLASS
class MicronucleiDataset(Dataset):
    
    def __init__(self, directory, mode="random", scale_factor=1.0, patch_size=256, stride=8, feature_size=384, edges=False, transform=None, gaussian=False):
        # Store parameters
        self.patch_size = patch_size
        self.stride = stride
        self.feature_size = feature_size
        self.mode = mode # in [random, fixed]
        self.edges = edges
        self.transform = transform
        self.shuffled = 0
        self.gaussian = gaussian
        self.directory = directory
        
        # Load images and annotations
        img_files = os.listdir(os.path.join(directory, 'images'))
        img_files= [file for file in img_files if not file.startswith('.')]
        
        self.images = {}
        for fname in tqdm(img_files):
            imid = fname.split('.')[0]
            im = read_image(os.path.join(directory, 'images', fname), scale=scale_factor)
            im = np.array((im - np.min(im))/(np.max(im) - np.min(im)), dtype="float32")
            mni = read_micronuclei_annotations(directory, fname)
            mnm = read_image(os.path.join(directory, 'mn_masks', fname.replace('.tif', '.png')), scale_factor) # expect micronuclei annotation in png format
            nuc = read_nuclei_masks(directory, fname) # nuclei mask in tif format
            self.images[imid] = {"image":im, "micro":mnm, "nuclei":nuc, "loc":mni}
            
            
        # Prepare data locations
        if self.mode == "random":
            self.randomize_patch_index()
        elif self.mode == "fixed":
            self.index_patches()
            self.transform = None
        else:
            assert False, "Incorrect mode"

        
    def randomize_patch_index(self): # This function is only applied to training set
        self.shuffled += 1
        #print("Randomized",self.shuffled,"times")
        self.index = []
        PS = self.patch_size
        
        df = pd.read_csv(os.path.join(self.directory.replace('/train', ''), 'metadata.csv'))
        for imid in self.images:
            # Generate random patch coordinates C
            H, W = self.images[imid]["image"].shape
            patches_per_image = (W // PS) * (H // PS)
            
            # Oversample more crops
            subset = df.loc[df.filenames == imid + '.tif', 'datasets'].iloc[0]
            if subset in ['HeLa', 'RPE1']:
                patches_per_image = patches_per_image * 7
            elif subset in ['BBBC039', 'mnfinder_train']:
                patches_per_image = patches_per_image * 5
                
            # print(f'{imid}: height - {H}, width - {W}, patches - {patches_per_image}')
            X = np.random.randint(0, W - PS, patches_per_image)
            Y = np.random.randint(0, H - PS, patches_per_image)
            C = np.stack((Y,X)).T
            A = {}

            # Micronuclei locations
            for k,r in self.images[imid]["loc"].iterrows():
                # Check whether the location r.x,r.y is covered by patches
                matches = np.where(np.logical_and( 
                            np.logical_and(C[:,0] < r.y, C[:,0] + PS > r.y),
                            np.logical_and(C[:,1] < r.x, C[:,1] + PS > r.x)
                ))
                matches = matches[0]
                if len(matches) > 0:
                    # Annotate all patches that cover the location
                    for m in matches:
                        try: A[m].append((r.y, r.x))
                        except: A[m] = [(r.y, r.x)]
                elif (r.y + PS < H or r.x + PS < W): # or statement might fit non-square images?
                    # If not covered, add a new patch that covers the location
                    if (r.y == 0 and r.x == 0):
                        extra = [[0, 0]]
                    elif (r.y != 0 and r.x == 0):
                        if (r.y + PS > H):
                            extra = [[H - PS, 0]]
                        else:
                            extra = [[np.random.randint(max(r.y - PS,0), r.y), 0]]
                    elif (r.y == 0 and r.x != 0):
                        if (r.x + PS > W):
                            extra = [[0, W - PS]]
                        else:
                            extra = [[0, np.random.randint(max(r.x - PS,0), r.x)]]
                    else:
                        if (r.y + PS < H and r.x + PS < W):
                            extra = [[np.random.randint(max(r.y - PS,0), r.y), np.random.randint(max(r.x - PS,0), r.x)]]
                        elif (r.y + PS < H and r.x + PS > W):
                            extra = [[np.random.randint(max(r.y - PS,0), r.y), W - PS]]
                        elif (r.y + PS > H and r.x + PS < W):
                            extra = [[H - PS, np.random.randint(max(r.x - PS,0), r.x)]]
                        elif (r.y + PS > H and r.x + PS > W):
                            extra = [[H - PS, W - PS]]
                    C = np.append(C, extra, axis=0)
                    A[C.shape[0]-1] = [(r.y, r.x)]

            # Put annotated patches in the index
            count = 0
            for k in A:
                self.index.append({"Image":imid, "coord": C[k,:].tolist(), "locs":A[k]})
                count += 1

            # Complete budget with non-annotated patches
            U = [x for x in range(patches_per_image) if x not in A]
            pointer = 0
            while count < patches_per_image:
                k = U[pointer]
                self.index.append({"Image":imid, "coord": C[k,:].tolist(), "locs":[]})
                pointer += 1
                count += 1
                

    def index_patches(self):
        self.index = []
        PS = self.patch_size
        
        for imid in self.images:
            # Generate regular grid of patch coordinates C
            H, W = self.images[imid]["image"].shape
            patches_per_image = (W // PS) * (H // PS)
            # print(f'{imid}: height - {H}, width - {W}, patches - {patches_per_image}')
            X = np.linspace(0, W - W % PS, W // PS + 1)
            Y = np.linspace(0, H - H % PS, H // PS + 1)
            X,Y = np.meshgrid(X[:-1],Y[:-1], indexing='ij')
            X = X.reshape((patches_per_image,))
            Y = Y.reshape((patches_per_image,))
            C = np.stack((Y,X)).T
            A = {}

            # Micronuclei locations
            for k,r in self.images[imid]["loc"].iterrows():
                # Find which patches cover the location r.x,r.y
                matches = np.where(np.logical_and( 
                            np.logical_and(C[:,0] < r.y, C[:,0] + PS > r.y),
                            np.logical_and(C[:,1] < r.x, C[:,1] + PS > r.x)
                ))
                matches = matches[0]
                if len(matches) > 0:
                    # Annotate all patches that cover the location
                    for m in matches:
                        try: A[m].append((r.y, r.x))
                        except: A[m] = [(r.y, r.x)]
                
                # comment ouf only for grid search purpose, remove commenting after grid search
                # else:
                #     print(f"{imid}: Micronuclei at ({r.y},{r.x}) is not covered by any patches")

            # Put all patches in the index
            for k in range(C.shape[0]):
                try:
                    self.index.append({"Image":imid, "coord": C[k,:].tolist(), "locs":A[k]})
                except:
                    self.index.append({"Image":imid, "coord": C[k,:].tolist(), "locs":[]})
                              
        
    def __len__(self):
        return len(self.index)
        
        
    def __getitem__(self, idx):
        item = self.index[idx]
        # print(item['Image'])
        
        if self.gaussian: # default: False
            height, width = np.random.normal(loc=self.patch_size, scale=20, size=2) # 20 is the best so far
            height, width = int(height), int(width)
            r,c = int(item["coord"][0]), int(item["coord"][1])
            crop = self.images[item["Image"]]["image"][r:r+height,c:c+width]
            mn_mask = self.images[item["Image"]]["micro"][r:r+height,c:c+width]
            n_mask = self.images[item["Image"]]["nuclei"][r:r+height,c:c+width]
            
            crop = patch_to_rgb(crop, self.edges)
            mask = torch.Tensor(np.concatenate(
                (mn_mask[np.newaxis,:,:], n_mask[np.newaxis,:,:]), axis=0
            ))

            # interpolate to Patch Size x Patch Size before concatention
            # interpolation messed up binary values sometime, remember to check
            crop = torch.nn.functional.interpolate(crop.unsqueeze(0), (self.patch_size, self.patch_size))
            mask = torch.nn.functional.interpolate(mask.unsqueeze(0), (self.patch_size, self.patch_size))
            
            crop = crop.squeeze(0)
            mask = mask.squeeze(0)
            
            # print(f'Crop shape: {crop.shape}')
            # print(f'Mask shape: {mask.shape}')
        
        # Crop patches out of the full image
        else: # original code
            PS = self.patch_size
            r,c = int(item["coord"][0]), int(item["coord"][1])
            crop = self.images[item["Image"]]["image"][r:r+PS,c:c+PS]
            mn_mask = self.images[item["Image"]]["micro"][r:r+PS,c:c+PS]
            n_mask = self.images[item["Image"]]["nuclei"][r:r+PS,c:c+PS]
            crop = patch_to_rgb(crop, self.edges)
            mask = torch.Tensor(np.concatenate(
                (mn_mask[np.newaxis,:,:], n_mask[np.newaxis,:,:]), axis=0
            ))
        
        if self.mode in ["random"] and self.transform is not None:
            crop, mask = self.transform(crop, mask)
            # mask = mask[0,:,:]
            
        return crop, mask
            
        


import sklearn.metrics
import wandb

import numpy as np
import pandas as pd

import scipy
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import skimage
import time


def compare_two_labels(label_model, label_gt, return_IoU_matrix, debug=False):
    
    # get number of detected nuclei
    nb_nuclei_gt = np.max(label_gt)
    nb_nuclei_model = np.max(label_model)
    
    # catch the case of an empty picture in model and gt
    if nb_nuclei_gt == 0 and nb_nuclei_model == 0:
        if(return_IoU_matrix):
            return [0, 0, 1, np.empty(0)]     
        else:
            return [0, 0, 1]
    
    # catch the case of empty picture in model
    if nb_nuclei_model == 0:
        if(return_IoU_matrix):
            return [0, nb_nuclei_gt, 0, np.empty(0)]     
        else:
            return [0, nb_nuclei_gt, 0]
    
    # catch the case of empty picture in gt
    if nb_nuclei_gt == 0:
        if(return_IoU_matrix):
            return [nb_nuclei_model, 0, 0, np.empty(0)]     
        else:
            return [nb_nuclei_model, 0, 0]
    
    # build IoU matrix
    IoUs = np.full((nb_nuclei_gt, nb_nuclei_model), -1.0, dtype = np.float32)

    # calculate IoU for each nucleus index_gt in GT and nucleus index_pred in prediction    
    # TODO improve runtime of this algorithm
    for index_gt in range(1,nb_nuclei_gt+1):

        nucleus_gt = label_gt == index_gt
        number_gt = np.sum(nucleus_gt)

        for index_model in range(1,nb_nuclei_model+1):
            
            if debug:
                print(index_gt, "/", index_model)
            
            nucleus_model = label_model == index_model 
            number_model = np.sum(nucleus_model)
            
            same_and_1 = np.sum((nucleus_gt == nucleus_model) * nucleus_gt)
            
            IoUs[index_gt-1,index_model-1] = same_and_1 / (number_gt + number_model - same_and_1)
    
    # get matches and errors
    detection_map = (IoUs > 0.5)
    nb_matches = np.sum(detection_map)

    detection_rate = IoUs * detection_map
    
    nb_overdetection = nb_nuclei_model - nb_matches
    nb_underdetection = nb_nuclei_gt - nb_matches
    
    mean_IoU = np.mean(np.sum(detection_rate, axis = 1))
    
    if(return_IoU_matrix):
        result = [nb_overdetection, nb_underdetection, mean_IoU, IoUs]
    else:
        result = [nb_overdetection, nb_underdetection, mean_IoU]
    return result

def measures_at(threshold, IOU):
    matches = IOU > threshold
    
    true_positives = np.sum(matches, axis=1) == 1   # Correct objects
    false_positives = np.sum(matches, axis=0) == 0  # Extra objects
    false_negatives = np.sum(matches, axis=1) == 0  # Missed objects
    
    assert np.all(np.less_equal(true_positives, 1))
    assert np.all(np.less_equal(false_positives, 1))
    assert np.all(np.less_equal(false_negatives, 1))
    
    TP, FP, FN = np.sum(true_positives), np.sum(false_positives), np.sum(false_negatives)
    
    f1 = 2*TP / (2*TP + FP + FN + 1e-9)

    prec = TP / (TP + FP)

    rec = TP / (TP + FN)

    return f1, prec, rec, TP, FP, FN


def segmentation_report(imid, predictions, gt, intersection_ratio=0.1, wandb_mode=False):
    # start = time.time()
    pi = skimage.morphology.label(predictions)
    gti = skimage.morphology.label(gt)
    nb_overdetection, nb_underdetection, mean_IoU, IoUs = compare_two_labels(pi, gti, True, False)
    if IoUs.size == 0:
        prec = 0.0
        rec = 0.0
        print(f'{imid}, Precision: {prec}, Recall: {rec}')
        if wandb_mode:
            wandb.log({'Precision':prec, 'Recall':rec})
    else:
        f1, prec, rec, TP, FP, FN = measures_at(intersection_ratio, IoUs)
        print(f'{imid}, Precision: {prec}, Recall: {rec}')
        if wandb_mode:
            wandb.log({'Precision':prec, 'Recall':rec})

def get_assignment(C, gt):
    # Map token predictions to pixels (multiply by 8)
    Y,X = list(C[0]*8), list(C[1]*8)
    C = np.asarray([(a,b) for a,b in zip(X,Y)])
    
    # Get true coordinates
    R = np.asarray(gt[["x","y"]])
    
    # Find nearest neighbor
    D = scipy.spatial.distance_matrix(C,R)
    assignment = np.argmin(D, axis=1)
    return assignment


def prediction_report(imid, probabilities, gt, threshold, output_dir):    
    ground_truth = np.zeros_like(probabilities)
    for k,r in gt.iterrows():
        a = r.y // 8
        b = r.x // 8
        ground_truth[a,b] = 1

    predictions = probabilities > threshold
    
    # Precision-recall curve

    GT = ground_truth.flatten()
    PRED = probabilities.flatten()

    plt.figure(figsize=(6,6))
    display = sklearn.metrics.PrecisionRecallDisplay.from_predictions(
        GT, PRED, name="Detector"#, plot_chance_level=True
    )
    _ = display.ax_.set_title("Precision-Recall curve")
    plt.savefig(f"{output_dir}/{imid}-prcurve.png")
    
    # Classification report
    
    report = sklearn.metrics.classification_report(GT, PRED > threshold)
    text_file = open(f"{output_dir}/{imid}-report.txt", "w")
    text_file.write(report)
    text_file.close()
    
    correct = predictions * ground_truth
    missing = ground_truth - correct
    extra = predictions - correct
    print("Total:",np.sum(ground_truth),"Correct:",np.sum(correct), "Missing:", np.sum(missing), "Extra:", np.sum(extra))
    
    results = {
        "probabilities": probabilities,
        "predictions": predictions,
        "ground_truth": ground_truth,
        "correct": correct,
        "missing": missing,
        "extra": extra
    }
    
    # Detailed report
    
    gt["Status"] = ""
    assignment = get_assignment(np.where(results["correct"]), gt)
    gt.loc[gt.index.isin(assignment), "Status"] = "correct"

    assignment = get_assignment(np.where(results["missing"]), gt)
    gt.loc[gt.index.isin(assignment), "Status"] = "missing" 
    
    gt.to_csv(f"{output_dir}/{imid}-details.csv", index=False)
    
    return results


def display_detections(im, imid, results, output_dir):
    # Show image
    fig, ax = plt.subplots(figsize=(30,30))
    ax.imshow(im)

    annotations = []

    # Display micronucleus boxes
    C = np.where(results["correct"])
    w,h = 16,16
    for i in range(len(C[0])):
        x1 = C[1][i]*8 - w
        y1 = C[0][i]*8 - h
        rect = patches.Rectangle((x1, y1), 2*w, 2*h, linewidth=1, edgecolor='gold', facecolor='none')
        ax.add_patch(rect)

    # Display micronucleus boxes
    C = np.where(results["missing"])
    w,h = 12,12
    for i in range(len(C[0])):
        x1 = C[1][i]*8 - w
        y1 = C[0][i]*8 - h
        rect = patches.Rectangle((x1, y1), 2*w, 2*h, linewidth=1, edgecolor='r', facecolor='none')
        ax.add_patch(rect)
        plt.text(x1, y1, i, color="r", fontsize="xx-large")
        annotations.append({"col":C[1][i]*8, "row":C[0][i]*8, "ID":i, "color":"red","question":"Missed?"})

    # Display micronucleus boxes
    C = np.where(results["extra"])
    w,h = 12,12
    for i in range(len(C[0])):
        x1 = C[1][i]*8 - w
        y1 = C[0][i]*8 - h
        rect = patches.Rectangle((x1, y1), 2*w, 2*h, linewidth=1, edgecolor='b', facecolor='none')
        ax.add_patch(rect)
        plt.text(x1, y1, i, color="b", fontsize="xx-large")
        annotations.append({"col":C[1][i]*8, "row":C[0][i]*8, "ID":i, "color":"blue","question":"Real?"})

    plt.axis('off')
    plt.show()
    plt.savefig(f"{output_dir}/{imid}-fig.png")
    
    df = pd.DataFrame(annotations)
    df["answer"] = ""
    df = df.sort_values(by=["ID","question"])
    df.to_csv(f"{output_dir}/{imid}-checks.csv", index=False)

